/*
 * processador_texto.java
 *
 * Created on 20 de Agosto de 2006, 18:19
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package protej;

import java.io.*;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.File.*;
import java.nio.CharBuffer;
import javax.swing.text.BadLocationException;
import javax.swing.text.Caret;

/**
 *
 * @author arnaldo
 */
public class processador_texto {
    
    private String arquivo_aberto = "";
    private String[] tokens = {" ", "\n", "\r", "\t", ".", "^", "<", ">", "/", "{", "}", "(", ")"}; // armazena tokens unitários
    private int total_tokens = tokens.length; // total de tokens unitários
    private javax.swing.JTextArea txt_texto; // armazena o objeto que carrega o texto
    private int cursor = 0;
    
    public boolean abre_arquivo(String nome_arquivo) {
    //
    // carrega um arquivo e exibe usando UTF-16
    //
        return abre_arquivo(nome_arquivo, "UTF-16LE");
    }
            
    public boolean abre_arquivo(String nome_arquivo, String codificacao) {
    //
    // carrega um arquivo e exibe ao usuário
    //
        try {
            // abrindo o arquivo
            FileInputStream alvo = new FileInputStream(nome_arquivo);
            InputStreamReader byte_para_char = new InputStreamReader(alvo, codificacao);
            int char_unicode;
            Reader leitor = new BufferedReader(byte_para_char);
            StringBuffer buffer = new StringBuffer();
            // validando Byte Order Maker (BOM)
            // nota importante: um editor hexadecimal mostra FFFE, mas como o arquivo 
            // está aberto como UTF-16LE, é preciso inverter os bytes para FEFF
            if (codificacao.equals("UTF-16LE") && leitor.read() != 0xFEFF) {
                System.out.println("Cabeçalho UTF-16LE ausente (0xfffe)");
                return false;
            }
            while ((char_unicode = leitor.read()) > -1) {
                buffer.append((char)char_unicode);
            }
            leitor.close();
            // mostrando na tela
            txt_texto.setText("");
            txt_texto.append(buffer.toString());
            txt_texto.select(0,0);
            arquivo_aberto = nome_arquivo;
            return true;
        } catch (IOException e) {
            e.printStackTrace();
            return false;
        }
    }  
   
    public void atualiza_cursor_externo() {
        //
        // faz o cursor externo apontar para onde o interno aponta
        //
        txt_texto.setCaretPosition(cursor);
    }
    
    public void atualiza_cursor_interno() {
        //
        // faz o cursor interno apontar para onde o externo aponta
        //
        cursor = txt_texto.getCaretPosition();
    }
    
    public String converte_para_string() {
        //
        // converte o texto editado para o formato string
        //
        return txt_texto.getText();
    }
    
    public boolean encontra_caracter(int direcao, String... caracteres_procurados) {
    // 
    // procura uma lista de caracteres e retorna true quando encontra algum
    //
        String caractere_lido;
        int i;
        int total_caracteres = caracteres_procurados.length;
        // definindo direcao
        if (direcao < 0) {
            direcao = -1;
        } else {
            direcao = 1;
        } 
        // buscando
        do {
            caractere_lido = le_caracter();
            for (i=0; i < total_caracteres; i++) {
                if (caracteres_procurados[i].equals(caractere_lido)) {
                    return true;
                }
            }
        } while (move(direcao) != 0);
        return false;
    }
    
    public boolean encontra_token(int direcao, String... tokens) {
    // 
    // procura uma lista de palavras e retorna true quando encontra alguma
    //
        int i;
        String token_lido;
        int total_tokens = tokens.length;
        // definindo direcao
        if (direcao < 0) {
            direcao = -1;
        } else {
            direcao = 1;
        } 
        // buscando
        do {
            token_lido = le_token();
            for (i=0; i < total_tokens; i++) {
                if (tokens[i].equals(token_lido)) {
                    return true;
                }
            }
        } while (move_token(direcao) != 0);
        return false;
    }

    public boolean encontra_texto(int direcao, String texto) {
    // 
    // procura um texto e retorna true quando encontra
    //
        int tamanho = texto.length();
        int posicao_atual = cursor;
        String texto_lido;
        if (direcao < 0) {
            direcao = -1;
        } else {
            direcao = 1;
        }
        do {
            texto_lido = le_caracter(tamanho);
            if (texto.equals(texto_lido)) {                
                return true;
            }
        } while (move(direcao) != 0);
        cursor = posicao_atual;
        return false;
    }

    public boolean encontra_naobranco(int direcao) {
    //
    // encontra a proxima ocorrência de caracter diferente de espaço e de tab
    //
        return encontra_naobranco(direcao, true);
    }
    
    public boolean encontra_naobranco(int direcao, boolean parar_enter) {
    //
    // encontra a proxima ocorrência de caracter diferente de espaço e de tab
    //
        boolean achou_branco;
        String caracter_lido;
        String[] caracteres_brancos = {" ", "\t", "\r", "", "\n"};
        int total_caracteres = 5;
        int i;
        if (parar_enter) {
            total_caracteres--;
        }
        // definindo direcao
        if (direcao < 0) {
            direcao = -1;
        } else {
            direcao = 1;
        } 
        // buscando
        do {
            caracter_lido = le_caracter();
            // nota: a comparação com vázio não está funcionando, mas foi mantida pela lógica da função
            achou_branco = false;
            for (i=0; i < total_caracteres; i++) {
                if (caracter_lido.equals(caracteres_brancos[i])) {
                    achou_branco = true;
                    break;
                }
            }
            if (! achou_branco) {
                return true;
            }
        } while (move(direcao) != 0);
        return false;
    }
        
    public void fecha_arquivo() {
    //
    // fecha sem salvar o arquivo em uso 
    //
        arquivo_aberto = "";
        cursor = 0;
        txt_texto.setText("");
    }
    
    public int insere_direita(String texto) {
    //
    // insere texto a direita do cursor
    // retorna 0 em caso de falha
    //
        int tamanho = texto.length();
        txt_texto.insert(texto, cursor);
        return tamanho;
    }
            
    public int insere_esquerda(String texto) {
    //
    // insere texto a esquerda do cursor
    // retorna 0 em caso de falha
    //
        int tamanho = texto.length();
        txt_texto.insert(texto, cursor);
        cursor += tamanho;
        return tamanho;
    }

    public String pega_arquivo_aberto() {
        return arquivo_aberto;
    }
    
    public String le_caracter() {
    //
    // lê o caracter a esquerda do cursor. Retorna vazio em caso de falha
    //
        return le_caracter(1);
    }

    public String le_caracter(int tamanho) {
    //
    // lê n caracteres a esquerda do cursor. Retorna vazio em caso de falha
    //
        String caracteres;
        int posicao_dot;
        try {
            caracteres = txt_texto.getText(cursor, tamanho);
        } catch (BadLocationException ex) {
            caracteres = "";
        }
        return caracteres;
    }
    
    public String le_caracter(int inicio, int fim) {
    //
    // lê o caracter da posição inicial até a final (não inclusive). Retorna vazio em caso de falha
    //
        String caracteres;
        try {
            caracteres = txt_texto.getText(inicio, fim - inicio);
        } catch (BadLocationException ex) {
            caracteres = "";
        }
        return caracteres;
    }
    
    public int le_natural() {
    //
    // retorna o numero lido ou -1 caso o cursor não esteja posicionado sobre um número
    //
        int numero = -1;
        int posicao_atual = le_posicao();
        char caracter = le_caracter().toCharArray()[0];
        if (caracter >= '0' && caracter <= '9') {
            numero = caracter - '0';
            while (move(1) == 1) {
                caracter = le_caracter().toCharArray()[0];
                if (caracter >= '0' && caracter <= '9') {
                    numero = numero * 10 + (caracter - '0');
                } else {
                    break;
                }
            }
        }
        cursor = posicao_atual;
        return numero;
    }
    
    public int le_posicao() {
    //
    // retorna a posicação do cursor
    //
        return cursor;
    }
    
    public String le_tag_xml() {
    //
    // retorna a tag xml na posição do cursor ou vazio
    //
        int posicao_inicial;
        String tag;
        if (! le_caracter().equals("<")) {
            return "";
        }
        posicao_inicial = le_posicao();
        move(1);
        encontra_naobranco(1);
        tag = le_token();
        if (tag.equals("/")) {
            move(1);
            tag += le_token();
        }
        encontra_caracter(1, ">", "\n");
        if (!le_caracter().equals(">")) {
            move_para_posicao(posicao_inicial);
            return "";
        }
        move_para_posicao(posicao_inicial);
        return tag;
    }
    
    public int le_tamanho_texto() {
    //
    // returna o tamanho do texto em caracteres
    //
        return txt_texto.getDocument().getLength();
    }
    
    public String le_token() {
    //
    // retorna um token a partir da posição cursor
    // tokens simples: espaço, lf, lf+cr, números, chaves, etc
    // tokens compostos: unicode-simples[unicode-simples]*
    //
        boolean fim_token = false;
        String caracter = "";
        String token = "";
        int i, j, posicao_atual = le_posicao();
        // lendo tokens compostos até encontrar um token unitário
        do {
            caracter = le_caracter();
            if (token_simples(caracter)) {
                fim_token = true;
            } else {
                token += caracter;
            }
        } while (move(1) == 1 && ! fim_token);
        // tokens simples
        if (token.equals("")) {
            token = caracter;
        }
        cursor = posicao_atual;
        return token;
    }
    
    public int move(int deslocamento) {
    //
    // move o cursor para a direita (ou esquerda) e retorna o número de bytes movidos 
    //
        boolean fim_arquivo;
        int posicao_atual, nova_posicao, tamanho_texto;
        posicao_atual = cursor;
        nova_posicao = posicao_atual + deslocamento;
        if (nova_posicao < 0) {
            cursor = 0;
        } 
        else {
            tamanho_texto = txt_texto.getDocument().getLength();
            if (nova_posicao <= tamanho_texto) {
                cursor = nova_posicao;
            } else {
                cursor = tamanho_texto;
            }
        }
        return cursor - posicao_atual;
    }
    
    public void move_inicio() {
    // 
    // move o cursor para a o inicio do documento (útil para acesso externo)
    //
        cursor = 0;
    }

    public int move_natural(int deslocamento) {
    //
    // pula um número natural e retorna o resultado
    //
        char caracter = le_caracter().toCharArray()[0];
        int movidos = 0;
        if (deslocamento > 0) {
            // mover um número para frente
            do {
                caracter = le_caracter().toCharArray()[0];
                if (caracter < '0' || caracter > '9') {
                    break;
                } else {
                    movidos = 1;
                }
            } while (move(1) == 1);
            return movidos;
        } else if (deslocamento < 0) {
            // mover um número para trás
            if (move(-1) != -1) {
                return 0;
            }
            do  {
                caracter = le_caracter().toCharArray()[0];
                if (caracter < '0' || caracter > '9') {
                    move(1);
                    break;
                } else {
                    movidos = -1;
                }
            } while (move(-1) == -1);
            return movidos;
        }
        return 0;
    }
    
    public boolean move_para_posicao(int posicao) {
    // 
    // move o cursor para a posição dada
    //
        int tamanho = txt_texto.getDocument().getLength();
        if (posicao < 0) {
            cursor = 0;
            return false;
        }
        if (posicao > tamanho) {
            cursor = tamanho;
            return false;
        }
        cursor = posicao;
        return true;
    }
    
    public int move_token(int deslocamento) {
    // 
    // move o cursor n tokens para a esquerda ou direita
    //
        String caracter;
        int direcao;
        int i = 0;
        boolean atual_simples = false;
        boolean anterior_simples = false;
        int tokens_movidos = 0;
        // definindo a direcao
        if (deslocamento < 0) {
            direcao = -1;
            deslocamento = -deslocamento;
        } else {
            direcao = 1;
        }
        // movendo
        atual_simples = token_simples(le_caracter());
        while (tokens_movidos < deslocamento) {
            if ((i = move(direcao)) == 0) {
                break;
            }
            anterior_simples = atual_simples;
            atual_simples = token_simples(le_caracter());
            if (atual_simples || anterior_simples) {
                tokens_movidos++;
            }
        }
        // indo para início da palavra quando direção é a esquerda
        if (direcao == -1) {
            if (! atual_simples && i != 0) {
                while (!token_simples(le_caracter())) {
                    if ((i = move(-1)) == 0) {
                        break;
                    }
                }
                if (i != 0) {
                    move(1);
                }
            }
            tokens_movidos = - tokens_movidos;
        }
        return tokens_movidos;
    }
    
    public processador_texto(javax.swing.JTextArea text_area) {
    // 
    // init da classe
    //
        txt_texto = text_area;
    }
    
    public int remove(int caracteres) {
    //
    // remove caracteres a direita do cursor e retornando o número de deletados
    // remove a esquerda quando offset < 0
    //
        int posicao_inicial, posicao_final;
        posicao_inicial =  cursor;
        move(caracteres);
        posicao_final = cursor;
        if (caracteres > 0) {
            txt_texto.select(posicao_inicial, posicao_final);
            txt_texto.replaceSelection("");
            cursor = posicao_inicial;
            return posicao_final - posicao_inicial;
        } else if (caracteres < 0 ) {
            txt_texto.select(posicao_final, posicao_inicial);
            txt_texto.replaceSelection("");
            cursor = posicao_final;
            return posicao_final - posicao_inicial;
        }
        return 0;
    }

    public int remove(int posicao_inicial, int posicao_final) {
    //
    // remove caracteres da posicao inicial ate a posicao final (não inclusive)
    //
        if (posicao_final <= posicao_inicial) {
            return 0;
        }
        txt_texto.select(posicao_inicial, posicao_final);
        txt_texto.replaceSelection("");
        cursor = posicao_inicial;
        return (posicao_final - posicao_final);
    }

    public int remove_natural() {
    //
    // remove uma seqüência de digitos e retorna 1 em caso de sucesso
    //
        char caracter;
        int posicao_final;
        int posicao_inicial = cursor;
        do {
             caracter = le_caracter().toCharArray()[0];
            if (caracter < '0' || caracter > '9') {
                break;
            }
        } while (move(1) == 1);
        posicao_final = cursor;
        if (posicao_final > posicao_inicial) {
            txt_texto.select(posicao_inicial, posicao_final);
            txt_texto.replaceSelection("");
            cursor = posicao_inicial;
            return 1;
        }
        return 0;
    }
    
    public int remove_tag() {
    //
    // remove uma tag xml e retorna 1 em caso de sucesso
    //
        int posicao_inicial;
        int removidos;
        String tag;
        if (! le_caracter().equals("<")) {
            return 0;
        }
        posicao_inicial = le_posicao();
        encontra_caracter(1, ">", "\n");
        if (!le_caracter().equals(">")) {
            move_para_posicao(posicao_inicial);
            return 0;
        }
        remove(posicao_inicial, le_posicao());
        remove(1);
        return 1;
    }
    
    public int remove_token(int tokens) {
    //
    // remove caracteres a direita do cursor e retornando o número de deletados
    // remove a esquerda quando offset < 0
    //
        int i, posicao_inicial, posicao_final;
        posicao_inicial = cursor;
        i = move_token(tokens);
        posicao_final = cursor;
        if (i > 0) {
            txt_texto.select(posicao_inicial, posicao_final);
            txt_texto.replaceSelection("");
            cursor = posicao_inicial;
        } else if (i < 0) {
            txt_texto.select(posicao_final, posicao_inicial);
            txt_texto.replaceSelection("");
            cursor = posicao_final;
        }
        return i;
    }

    public boolean salva_arquivo(String nome_arquivo) {
        return salva_arquivo(nome_arquivo, "UTF-16LE");
    }
    
    public boolean salva_arquivo(String nome_arquivo, String codificacao) {
    //
    // carrega um arquivo e exibe ao usuário
    //
        try {
            if (nome_arquivo.equals("")) {
                if (arquivo_aberto.equals("")) {
                    return false;
                }
                nome_arquivo = arquivo_aberto;
            } else {
                arquivo_aberto = nome_arquivo;
            }
            FileOutputStream alvo = new FileOutputStream(nome_arquivo);
            Writer escritor = new OutputStreamWriter(alvo, codificacao);
            if (codificacao.equals("UTF-16LE")) {
                escritor.write(0xFEFF); // está invertido devido ao uso de little endian
            }
            escritor.write(txt_texto.getText());
            escritor.close();
            return true;
        } catch (IOException e) {
            e.printStackTrace();
            return false;
        }
    }

    
    private boolean token_simples(String caracter) {
    //
    // indica se o token lido é unitário
    //
        int i;
        for (i=0; i < total_tokens; i++) {
            if (caracter.equals(tokens[i])) {
                return true;
            } // if
        } // for
        return false;
    } // function
} // class
